/* -------------------------------------------------------------------	*/
/* Copyright (C) 2002 Robert R. Bruner.  All rights reserved.		*/
/* Contact the author at						*/
/*                    rrb@math.wayne.edu				*/
/*                    Dept of Mathematics				*/
/*                    Wayne State University				*/
/*                    Detroit, Michigan  48202				*/
/*                    USA						*/
/*									*/
/* Version 1.7								*/
/* -------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>

#include <steen.h>

/* WARNING: Mult will stay within the subalgebra defined by N_ALG (see  */
/* the comments at the start of Mult) but none of the other routines    */
/* are that smart at present.  EG when N_ALG == 2, the algebra is 8     */
/* dimensional, and has, for example, only one element of degree 4,     */
/* namely (1,1).  However, Seqno, first_op, advance_op, etc will all    */
/* act as if (4) were also a legal operation.                           */

static int DEGREE[N_ALG] =
#if N_ALG == 1
     {1};
#endif
#if N_ALG == 2
     {1,3};
#endif
#if N_ALG == 3
     {1,3,7};
#endif
#if N_ALG == 4
     {1,3,7,15};
#endif
#if N_ALG == 5
     {1,3,7,15,31};
#endif
#if N_ALG == 6
     {1,3,7,15,31,63};
#endif
#if N_ALG == 7
     {1,3,7,15,31,63,127};
#endif
/* --------------------------------------------  degree     ---------- */
int degree(Op) Internal_Op Op;
{ int i,d=0;

for (i=0; i < N_ALG; i++)
   d += Op[i] * DEGREE[i];
return(d);
}
/* --------------------------------------------  fprint_op   ---------- */
void fprint_op(fp,Op) FILE *fp; Internal_Op Op;
{ register int i,max;
  for (max=N_ALG; Op[--max]==0 && (max >0); );
  fprintf(fp,"(%d",Op[0]);
  for (i=1; i<=max; i++) fprintf(fp,",%d",Op[i]);
  fprintf(fp,")");
}
/* --------------------------------------------  print_op   ---------- */
void print_op(Op) Internal_Op Op;
{ fprint_op(stdout,Op); }

/* --------------------------------------------  first_op      ---------- */
int first_op(Op,deg) Internal_Op Op; int deg;
{ /* set the operation Op to the first Steenrod operation in degree    */
  /* deg, in (reverse) lexicographic order.  (Reverse because read     */
  /* right to left, not left to right: ie (3,0,0) < (0,1,0).           */
  /* Returns the sequence number of the first op, and sets Op to that  */
  /* operation, if there are any.  Otherwise returns -1                */

int i;
Op[0] = deg;
for (i=1; i<N_ALG; i++) Op[i] = 0;
return(0);
}

/* ---------------------------------------  advance_op      ---------- */
int advance_op(Op) Internal_Op Op; {
/*  Advance Op to the lexically next operation in the same degree.          */
/*  (See first_op for comments on the order.) Return 1 if Op can be         */
/*  advanced, 0 if it is the last one in this degree.                       */

int i;
for (i=1; i<N_ALG; i++)
   if (Op[0] >= DEGREE[i]) { Op[0] -= DEGREE[i];
                             Op[i]++;
                             return(1);  }
   else { Op[0] += Op[i]*DEGREE[i];
          Op[i] = 0;  }
return(0);
}

/* --------------------------------------------  next_op      ---------- */
int next_op(Op,seqno) Internal_Op Op; int seqno; {
/*  Advance Op to the lexically next operation in the same degree.          */
/*  (See first_op for comments on the order.) Return it's sequence number,  */
/*  seqno+1, if Op can be advanced.  Return -1 if it is the last one in     */
/*  this degree.                                                            */

int i;
for (i=1; i<N_ALG; i++)
   if (Op[0] >= DEGREE[i]) { Op[0] -= DEGREE[i];
                             Op[i]++;
                             return(seqno+1);  }
   else { Op[0] += Op[i]*DEGREE[i];
          Op[i] = 0;  }
return(-1);
}
/* --------------------------------------------  adv_op ------------------ */
int adv_op(Op,k) Internal_Op Op; int k; {
/*  Advance Op k steps in lexicographic order (k >= 0 here). Return 1 if   */
/* Op can be advanced, 0 if it would go past the end.                      */
int i,r3;

r3 = Op[0]/3;
if (k <= r3) { Op[0] -= 3*k;
               Op[1] += k;
               return(1);
             }
Op[0] += 3*Op[1];
Op[1] = 0;
k -= r3;

for (i=2; i < N_ALG; i++)
  if (Op[0] >= DEGREE[i]) { Op[0] -= DEGREE[i];
                            Op[i]++;
                            k--;
                            r3 = Op[0]/3;
                            if (k <= r3) { Op[0] -= 3*k;
                                           Op[1] += k;
                                           return(1);
                                         }
                            k -= r3;
                            i = 1;
                          }
  else { Op[0] += Op[i] * DEGREE[i];
         Op[i] = 0;
       }

return(0);
}





/* --------------------------------------------  d (= dim) function  ----- */
unsigned d(k,n) unsigned k; int n; {
/* d(k,n) is the number of sequences (r1,r2,...,rk) whose degree is n,     */
/* where the degree is r1 + 3*r2 +7*r3 +15*r4 +...                         */
unsigned sum;

if (k<=2)  { if (n<0) return(0); else return(1+n/3);}

for (sum=0,k--; n >= 0; n -= DEGREE[k])  sum += d(k,n);

return(sum);
}

/* --------------------------------------------  Seqno    ---------------- */
unsigned Seqno(Op) Internal_Op Op; {
/*  Return the number of operations in this degree which come before Op    */
/*  in reverse lexicographic order.                                        */
unsigned s=0;
int i,deg;
unsigned k;

for (deg=0,i=0; i<N_ALG; i++) deg += Op[i]*DEGREE[i];

for (k=N_ALG; k-- > 2; )
   for (i=0; i < Op[k]; i++) { s += d(k,deg);
                               deg -= DEGREE[k];  }
return(s + Op[1]);
}


/*  -----  routines used by mult and its global variables---------------   */

/* global variables */

int T[N_ALG+1],           /* hold terms of R * S, offset by one:           */
                          /*    T = (t1,t2,...,t[N]) (wasting T[0], but    */
                          /*    making the programming cleaner             */
    X[N_ALG+1][N_ALG+1],  /* matrix used to compute sum                    */
    N,maxrow,maxcol;      /* max nonzero diag, row and col                 */

/*  --------------------------------------------------------------------   */
int nextrow(row) int row;
{
                /* Try to advance row r, changing row 0 and col 0          */
                /* appropriately.  If successful, return 1.  If not,       */
                /* return 0 and reset row r to the start.                  */
int col, done=0;

for (col=1; (!done) && (col <= maxcol); col++)

    if (X[0][col] && (X[row][0] >> col))
      {                  /* can advance (row,col)                */
       X[row][col]++;
       X[row][0] -= 1 << col;
       X[0][col]--;
       done = 1;
      }
   else 
      {                /* can't advance it, so reset it.         */
       X[0][col] += X[row][col];
       X[row][0] += X[row][col] << col;
       X[row][col] = 0;
      }
return(done);
}

/*  --------------------------------------------------------------------   */
int next()
{
int row, done = 0;

for (row=1; (!done) && (row <= maxrow); row++)
  done = nextrow(row);
return(done);
}


/*  --------------------------------------------------------------------   */

int test()
{
int ok=1;
int diag, col, row, collim;

    for (diag=N; ok && (diag > 0); diag--)   
     { row = (maxrow < diag) ? maxrow : diag;
       T[diag] = X[row][col = diag - row];
       collim = (maxcol < diag) ? maxcol : diag;
       for (col++,row--; ok && (col <= collim); col++,row--)
         if (T[diag] & X[row][col]) ok = 0;
          else T[diag] += X[row][col];
     }
return(ok);
}

/* -------------------------------- mult      ---------------------------- */
int mult(R,S,x)   /* multiply two elements of the Steenrod algebra and add  */
                 /* the result to the operation in the TERM x.  We         */
                 /* assume they lie in the subalgebra generated by         */
                 /* Sq(2**i) for i = 0 to N_ALG - 1. Write N for N_ALG     */
		 /* in the rest of these coments.                          */

                 /* The vector R = (r0, r1, ..., r[N-1]) represents the    */
                 /* Milnor dual to the product of xi sub i to the r[i-1],  */
                 /* so must satisfy   0 <= r[N-i] < 2^i.  In particular,   */
                 /* r[0] < 2^N, so sq(2^{N-1}) is the highest              */
                 /* indecomposable Steenrod operation in the subalgebra.   */

/* The matrices X are ordered lexicographically by rows, with the first    */
/* row the least significant.  Rows are ordered lexicographically with     */
/* the column 1 entry the least significant.  The matrices are produced    */
/* in order.                                                               */

int *R, *S;
TERM *x;

{
int i, j, seq, notdone;

/* create initial matrix and set variables maxrow, maxcol, N and T         */

maxrow = maxcol = 1;
for (i=0; i < N_ALG; i++)
   { if ((X[0][i+1] = S[i])) maxcol = i+1;
     if ((X[i+1][0] = R[i])) maxrow = i+1;
     for (j=1; j <= N_ALG; j++) X[i+1][j] = 0;
   }
X[0][0] =0;
N = (N_ALG <= maxrow + maxcol) ? N_ALG : maxrow + maxcol;
for (i=N; i++ < N_ALG; ) T[i] = 0;



/* main loop  */

for (notdone=1; notdone; notdone=next())

  if (test())
  { 
    seq = Seqno(T+1);
    x->op[seq/8] ^= (0x80) >> (seq%8);
  }

return(0);
}


/* -----------------------------------------------------------------     */

/* ----------------- first_Coproduct ---------------------------------- */

int first_Coproduct(Op,Op1,Op2) Internal_Op Op,Op1,Op2;
{
int i;

for (i=0; i< N_ALG; i++)
  { Op1[i] = 0;
    Op2[i] = Op[i];
  }
return(1);
}

/* ----------------- next_Coproduct ---------------------------------- */

int next_Coproduct(Op,Op1,Op2) Internal_Op Op,Op1,Op2;
{
int i;

for (i=0; i< N_ALG; i++)
  if (Op1[i] < Op[i]) { Op1[i]++; Op2[i]--; return(1); }
   else { Op1[i] = 0; Op2[i] = Op[i]; }
return(0);
}


/* ----------------- halve (for Sq0) --------------------------------- */

int halve(Op,hOp) Internal_Op Op,hOp;
{
int i;
int odd;

odd = 0;

for (i=0; i < N_ALG; i++)
{ odd |= Op[i];
  hOp[i] = ((int)Op[i])/2;
}
return(!(1 & odd));
}

